@@ -10,7 +10,8 @@
   keytype <- switch(EXPR = centralID,
                     "EG" = "ENTREZID",
                     "TAIR" = "TAIR",
-                    "ORF" = "ORF")
+                    "ORF" = "ORF",
+                    "GID" = "GID")
   keytype
 }
 
@@ -726,12 +727,6 @@
   cols
 }
 
-.testIfKeysAreOfProposedKeytype <- function(x, keys, keytype){
-  ktKeys = keys(x, keytype=keytype)
-  if(!(any(ktKeys %in% keys))){
-    stop("None of the keys entered are valid keys for the keytype specified.")
-  }
-}
 
 ## helper looks at cols and trys to warn users about many:1 relationships
 .warnAboutManyToOneRelationships <- function(cols){
@@ -743,12 +738,12 @@
         msg = paste("You have selected the following columns that can have a many to one relationship with the primary key: ", paste(problemCols,collapse=", "),". Because you have selected more than a few such columns there is a risk that this selection may balloon up into a very large result as the number of rows returned multiplies accordingly. To experience smaller/more manageable results and faster retrieval times, you might want to consider selecting these columns separately.") 
         warning(paste(strwrap(msg, exdent=2), collapse="\n"),
                 immediate.=TRUE, call.=FALSE)
-    }    
+    }
 }
 
 
 ## the core of the select method for GO org and chip packages.
-.select <- function(x, keys=NULL, cols=NULL, keytype, jointype) {
+.legacySelect <- function(x, keys=NULL, cols=NULL, keytype, jointype) {
   ## if asked for what they have, just return that.
   if(all(cols %in% keytype)  && length(cols)==1){
     res <- data.frame(keys=keys)
@@ -756,11 +751,9 @@
     return(res)
   }
   if(is.null(keys)) keys <- keys(x) ## if no keys provided: use them all
-  if(is.null(cols)) cols <- cols(x) ## if no cols provided: use them all
-  ## check that the keytype matches the keys
-  .testIfKeysAreOfProposedKeytype(x, keys, keytype)
-
-  ## call .simplifyCols to ensure we use same colnames as cols()
+  if(is.null(cols)) cols <- columns(x) ## if no cols provided: use them all
+  
+  ## call .simplifyCols to ensure we use same colnames as columns()
   cols <- .simplifyCols(x, cols)
   ## keytype <- .swapSymbolExceptions(x, keytype)
   keytype <- .simplifyCols(x, keytype)
@@ -820,8 +813,136 @@
   res
 }
 
+
+## This just needs to generate a simple query and then return the
+## results with no _id's
+## A simple strategy will work EXCEPT for the weird case where I need
+## to pull out multiple fields at once. (like GO)
+
+## Do I need to pay attention to the order??? - it seems that I don't...
+.appendGenesToTabs <- function(tabs){
+    names <- names(tabs)
+    tabs <- c("genes",tabs)
+    names(tabs) <- c("GID", names)
+    tabs
+}
+
+.noSchemaSelect <- function(x, keys=NULL, cols=NULL, keytype){
+
+    ## 1st pool all the fields we need to extract
+    fields <- unique(c(cols, keytype))
+    ## Then get the tables to go with each one.
+    tabs <- sapply(fields, .deriveTableNameFromField, x=x)
+    ## make fully qualified fields of these tabs (the ones we want to extract)
+    f.fields <- paste(tabs, fields, sep=".")
+
+    ## if it's a Chip package, attach and point to org package for x
+    if(class(x)=="ChipDb"){
+        y <- x
+        ## then flip to using the org package, and actually attach to that.
+        x <- .getOrgPkg(x)
+        try(.attachDB(x,y), silent=TRUE) ## not a disaster if we fail        
+        ## Also make sure we include the genes table (only needed for the join)
+        if(!("genes" %in% tabs)) tabs <- .appendGenesToTabs(tabs)
+    }
+    
+    ## Make non-redundant list of tables to visit
+    nrTabs <- unique(tabs)
+    ## Now join to each table
+    for(i in seq_along(nrTabs)){
+        if(i==1){
+            sql <- paste("SELECT ",paste(f.fields, collapse=","),
+                         " FROM",tabs[1])
+        }else{
+            ## IF we see c.probes in nrTabs[i], it means we have to
+            ## use gene_id instead.
+            if("c.probes" %in% nrTabs[i]){
+                sql <- c(sql, paste("LEFT JOIN ",nrTabs[i],"USING (GID)"))
+            }else{
+                sql <- c(sql, paste("LEFT JOIN ",nrTabs[i],"USING (_id)"))
+            }
+        }
+    }
+    sql <- paste(sql, collapse=" ")
+    ## add the where clause
+    strKeys <- paste0('"',keys,'"',collapse=",")
+    fullKeytype <- tabs[names(tabs)==keytype]
+    fullKeytype <- paste(fullKeytype, names(fullKeytype), sep=".") 
+    where <- paste("WHERE ",fullKeytype,"in (",strKeys,")" )
+    sql <- paste(sql, where)
+    ## then call that
+    res <- dbQuery(dbConn(x), sql)    
+    ## cleanup and re-organize
+    .resort(res, keys, jointype=keytype, fields)
+}
+
+###############################################################################
+## HELPERS for argument validation {in select() and keys()}
+
+.isSingleString <- function(x){
+  is.atomic(x) && length(x) == 1L && is.character(x)
+}
+
+.testForValidKeytype <- function(x, keytype){
+  if(!.isSingleString(keytype)){
+      stop("'keytype' must be a a single string")
+  }
+  pkts <- keytypes(x)
+  if(!(keytype %in% pkts)){
+      msg <- paste0("Invalid keytype: ",keytype,". Please use the keytypes method to see a listing of valid arguments.")
+      stop(msg)
+  }
+}
+
+.testForValidCols <- function(x, cols){
+  if (!is.character(cols)){
+      stop("'columns' must be a character vector")
+  }
+  pcols <- columns(x)
+  if(!all(cols %in% pcols) && !is.null(cols)){
+      badCols <- cols[!(cols %in% pcols)]
+      msg <- paste0("Invalid columns: ",paste(badCols, collapse=","),". Please use the columns method to see a listing of valid arguments.")
+      stop(msg)
+  }
+}
+
+.testForValidKeys <- function(x, keys, keytype){
+  if (!is.character(keys)){
+      stop("'keys' must be a character vector")
+  }
+  ktKeys <- keys(x, keytype)
+  if(!(any(ktKeys %in% keys))){
+      msg <- paste0("None of the keys entered are valid keys for '",keytype,
+         "'. Please use the keys method to see a listing of valid arguments.")
+      stop(msg)
+  }
+}
+
+.testSelectArgs <- function(x, keys, cols, keytype){
+    .testForValidKeytype(x, keytype)
+    .testForValidCols(x, cols)
+    .testForValidKeys(x, keys, keytype)
+}
+
+
+## general select function
+.select <- function(x, keys=NULL, cols=NULL, keytype, jointype){
+    ## Some argument checking
+    .testSelectArgs(x, keys=keys, cols=cols, keytype=keytype)
+    ## Now get the schema
+    schema <- metadata(x)[metadata(x)$name=="DBSCHEMA",]$value
+    if(schema=="NOSCHEMA_DB" || schema=="NOCHIPSCHEMA_DB"){
+        .noSchemaSelect(x, keys, cols, keytype)
+    }else{
+        .legacySelect(x, keys, cols, keytype, jointype)
+    }
+}
+
+
+
+
 ## Helper for setting the jointype to an appropriate default
-.chooseJointType  <- function(x){
+.chooseJoinType  <- function(x){
   if(.getCentralID(x) == "ORF"){
     jointype <- "systematic_name"
   }else{
@@ -831,28 +952,52 @@
 }
 
 
+#######################################################################
+## So an issue was that keytype was not defined carefully enough and
+## was getting substituted when cols arg was used in conjunction with
+## no keytype specified...  So when there is not keytype, the user
+## must tell us which one it is OR we have to somehow "know"
+
+## TO FIX: I need to guess the keytype and pass it along as some other
+## name into ...
+## Hack: I will pass it in as kt
+## And halfway through this hack, it seems that if there is an
+## argument passed in AFTER your other ones that the problem I was
+## having magically clears up???  WTH?
+
 
 setMethod("select", "OrgDb",
-    function(x, keys, cols, keytype) {
-          if (missing(keytype)){
-            keytype <- .chooseCentralOrgPkgSymbol(x)
+          function(x, keys, columns, keytype, ...) {
+##               if (missing(keytype)){
+##                   keytype <- .chooseCentralOrgPkgSymbol(x)
+##               }
+              kt <- .chooseCentralOrgPkgSymbol(x)
+              jointype <- .chooseJoinType(x)
+
+              .selectWarnJT(x, keys, columns, keytype, jointype=jointype,
+                            kt=kt, ...)
+              ## put back following line after 2.13 releases
+ ##              .select(x, keys, columns, keytype, jointype=jointype)
           }
-          jointype <- .chooseJointType(x)
-          .select(x, keys, cols, keytype, jointype=jointype)
-        }
 )
 
 setMethod("select", "ChipDb",
-    function(x, keys, cols, keytype){
-          if (missing(keytype)) keytype <- "PROBEID"
-          .select(x, keys, cols, keytype, jointype="probe_id")
-        }
+    function(x, keys, columns, keytype, ...){
+##         if (missing(keytype)) keytype <- "PROBEID"
+        kt <- "PROBEID"
+        .selectWarnJT(x, keys, columns, keytype, jointype="probe_id",
+                      kt=kt, ...)
+        ## put back following line after 2.13 releases
+        ##           .select(x, keys, columns, keytype, jointype="probe_id")
+    }
 )
 
 setMethod("select", "GODb",
-    function(x, keys, cols, keytype){
+    function(x, keys, columns, keytype, ...){
           if (missing(keytype)) keytype <- "GOID"
-          .select(x, keys, cols, keytype, jointype="go_id")
+          .selectWarnJT(x, keys, columns, keytype, jointype="go_id", ...)
+          ## put back following line after 2.13 releases
+##           .select(x, keys, columns, keytype, jointype="go_id")
         }
 )
 
@@ -861,10 +1006,9 @@ setMethod("select", "GODb",
 
 
 
-
 #############################
 ## Internally we want to reconstruct these guys so we can merge() on them
-## c <- cols(GO.db)[7]
+## c <- columns(GO.db)[7]
 ## prefix = "GO"
 ## foo = paste0(prefix,c)
 ## bar = eval(parse(text=foo))
@@ -876,7 +1020,7 @@ setMethod("select", "GODb",
 ## just the table names, or it might be a list of mappings
 
 
-.cols <- function(x, baseType){
+.legacyCols <- function(x, baseType){
   ## cols <- .makeColAbbrs(x)
   cols <- names(.defineTables(x))
   if(!missing(baseType)){
@@ -891,19 +1035,50 @@ setMethod("select", "GODb",
   unique(cols)
 }
 
+## These helpers will go into the DB and extract col values for newer
+## NOSCHEMA_DB's
 
-setMethod("cols", "OrgDb",
+.getDataTables <- function(con){
+  tables <- dbListTables(con)
+  tables[!tables %in% c("metadata","map_metadata","map_counts")]
+}
+
+.noSchemaCols <- function(x){
+  if(class(x)=="ChipDb"){
+      y <- x ## The old switcheroo
+      x <- .getOrgPkg(x)  
+  }
+  con <- dbConn(x)
+  tables <- .getDataTables(con)
+  cols <- unique(unlist(sapply(tables, FUN=dbListFields, con=con)))
+  cols <- cols[!cols %in% "_id"]
+  if(exists('y')){cols <- c("PROBEID", cols)}
+  cols
+}
+
+## general .cols function
+.cols <- function(x, baseType){
+    schema <- metadata(x)[metadata(x)$name=="DBSCHEMA",]$value
+    if(schema=="NOSCHEMA_DB" || schema=="NOCHIPSCHEMA_DB"){
+        .noSchemaCols(x)
+    }else{
+        .legacyCols(x, baseType)
+    }
+}
+
+
+setMethod("columns", "OrgDb",
     function(x){
       baseType <- .chooseCentralOrgPkgSymbol(x)
       .cols(x, baseType)
     }
 )
 
-setMethod("cols", "ChipDb",
+setMethod("columns", "ChipDb",
     function(x) .cols(x, baseType="PROBEID")
 )
 
-setMethod("cols", "GODb",
+setMethod("columns", "GODb",
     function(x) .cols(x) ## does not have a missing baseType
 )
 
@@ -940,10 +1115,11 @@ setMethod("cols", "GODb",
 }
 
 
-.makeKeytypeChoice <- function(x, keytype){
+.legacyKeys <- function(x, keytype){
   ## have to swap keytype
   ## keytype <- .swapSymbolExceptions(x, keytype)
   keytype <- .simplifyCols(x, keytype)
+  
   ## Some org packages may have entrez genes in weird places...
   centralID <- .getCentralID(x)
   EGgeneTable <- character()
@@ -973,35 +1149,189 @@ setMethod("cols", "GODb",
                     "SELECT DISTINCT probe_id FROM probes", 1L),
                   .queryForKeys(x, keytype))
   }
-  res[!is.na(res)]
+  if(class(x) == "GODb"){
+    res <- switch(EXPR = keytype,
+                  "GOID" =  dbQuery(dbConn(x),
+                    "SELECT DISTINCT go_id FROM go_term", 1L),
+                  .queryForKeys(x, keytype))
+  }
+  as.character(res[!is.na(res)])
 }
 
 
-## TODO: swap initial SQL query for an Lkeys() call for OrgDb and ChipDb??
+## special functions for newer NOSCHEMA_DB's
+.deriveTableNameFromField <- function(field, x){
+    if(class(x)=="ChipDb"){
+        y <- x ## Switcheroo
+        x <- .getOrgPkg(x)
+        try(.attachDB(x,y), silent=TRUE) ## not a disaster if we fail
+    }
+    con <- dbConn(x)
+    tables <- .getDataTables(con)
+    if(exists("y")){
+        tables <- c("c.probes", tables)
+    }
+    colTabs <- sapply(tables, FUN=dbListFields, con=con)
+    m <- unlist2(sapply(colTabs, match, field))  ## cannot ever be repeated
+    tab <- names(m)[!is.na(m)]
+    if(length(tab)!=1){stop("Two fields in the source DB have the same name.")}
+    tab
+}
+
+.noSchemaKeys <- function(x, keytype){
+    tab <- .deriveTableNameFromField(field=keytype, x)
+    ## So now we know table name (tab) and field (keytype)
+    if(class(x)=="ChipDb"){
+        y <- x ## Switcheroo
+        x <- .getOrgPkg(x)
+        try(.attachDB(x,y), silent=TRUE) ## not a disaster if we fail
+    }
+    sql <- paste("SELECT",keytype,"FROM",tab)
+    res <- dbQuery(dbConn(x), sql, 1L)
+    as.character(res[!is.na(res)])
+}
+
+
+## general keys function
+.keys <- function(x, keytype){
+    .testForValidKeytype(x, keytype)
+    schema <- metadata(x)[metadata(x)$name=="DBSCHEMA",]$value
+    if(schema=="NOSCHEMA_DB" || schema=="NOCHIPSCHEMA_DB"){
+        .noSchemaKeys(x, keytype)
+    }else{
+        .legacyKeys(x, keytype)
+    }
+}
+
+
+####################################################################
+## So the new idea is that each place where I want to "enhance" keys,
+## I should just be able to use a helper to wrap up the actual keys
+## method...
+
+
+## And we need a master helper to tie it all together
+smartKeys <-
+    function(x, keytype, ..., pattern, column, fuzzy=FALSE, FUN)
+{
+    ## check args, then...
+
+    ## FUN is the base keys method
+    .keys <- FUN
+    
+    ## So 1st we need helpers for other "keys" situations
+    ## keys0 is for when we have a pattern we want to match in the keys
+    .keys0 <- function(x, keytype, ..., pattern, fuzzy=FALSE)
+        {   ## assumes 'pattern' present
+            FUN <- if (fuzzy) agrep else grep
+            FUN(pattern, .keys(x, keytype), value=TRUE, ...)
+        }
+    ## keys1 is for when we have a column but no pattern
+    ## so we want to filter by column
+    .keys1 <- function(x, keytype, ..., column)
+        {   ## column acts as filter
+            k <- suppressWarnings(select(x, as.character(.keys(x, keytype)),
+                                         column, keytype))
+            k[[keytype]][ !is.na(k[[column]]) ]
+        }
+    ## keys2 is for when we have a column, and a pattern to match on that
+    ## column, and we want all the keys of a particular keytype that match
+    ## that column.
+    .keys2 <- function(x, keytype, ..., pattern, column, fuzzy=FALSE)
+        {   ## assumes 'pattern', 'column' present
+            FUN <- if (fuzzy) agrep else grep
+            k <- suppressWarnings(select(x, as.character(.keys(x, keytype)),
+                                         column, keytype))
+            k[[keytype]][ FUN(pattern, k[[column]], ...) ]
+        }
+
+    ## Now decide which function to call...
+    if (missing(pattern) && missing(column))
+        k <- .keys(x, keytype)
+    else if (missing(column))
+        k <- .keys0(x, keytype, ..., pattern=pattern, fuzzy=fuzzy)
+    else if (missing(pattern))
+        k <- .keys1(x, keytype, ..., column=column)
+    else
+        k <- .keys2(x, keytype, ..., pattern=pattern, column=column,
+                    fuzzy=fuzzy)
+    
+    unique(k)
+}
+
+
+
+## TODO: don't fail to document all the new arguments (pattern, column and fuzzy)
 setMethod("keys", "OrgDb",
-    function(x, keytype){
+    function(x, keytype, ...){
       if(missing(keytype)){
         keytype <- .chooseCentralOrgPkgSymbol(x)
       }
-      .makeKeytypeChoice(x, keytype)
-    }
+      smartKeys(x=x, keytype=keytype, ..., FUN=AnnotationDbi:::.keys)
+  }
 )
 
 setMethod("keys", "ChipDb",
-    function(x, keytype){
+    function(x, keytype, ...){
       if(missing(keytype)) keytype <- "PROBEID"
-      .makeKeytypeChoice(x, keytype)
-    }
+      smartKeys(x=x, keytype=keytype, ..., FUN=AnnotationDbi:::.keys)
+  }
 )
 
 setMethod("keys", "GODb",
-    function(x, keytype){
+    function(x, keytype, ...){
       if(missing(keytype)) keytype <- "GOID"
-      dbQuery(dbConn(x), "SELECT go_id FROM go_term", 1L)
-    }
+      smartKeys(x=x, keytype=keytype, ..., FUN=AnnotationDbi:::.keys)
+  }
 )
 
 
+## new uses for keys:
+## now TERM is a real key? (TODO: someone tell the keytypes)
+## head(keys(GO.db, keytype="TERM"))
+
+
+
+## get TERM keys that match a particular pattern
+## head(keys(GO.db, keytype="TERM", pattern="mitochondrion"))
+
+## get GOIDs where a TERM exists.
+## head(keys(GO.db, keytype="GOID", column="TERM"))
+
+
+
+## get keys of type GOID that go with a pattern match in TERM
+## head(keys(GO.db, keytype="GOID", pattern="mitochondrion", column="TERM"))
+## select(GO.db, keys =head(keys(GO.db, keytype="GOID", pattern="mitochondrion", column="TERM")), cols=c("GOID","TERM"))
+
+
+## do the above but use fuzzy matching
+## head(keys(GO.db, keytype="GOID", pattern="mitochondrion", column="TERM", fuzzy=TRUE))
+## select(GO.db, keys = head(keys(GO.db, keytype="GOID", pattern="mitochondrion", column="TERM", fuzzy=TRUE)), cols=c("GOID","TERM"))
+
+
+
+## Can just get keys (straight up)
+## head(keys(org.Hs.eg.db, keytype="SYMBOL"))
+
+## keys1 situation works fine (and smartKeys is called twice.)
+## Can filter by column (only return keys where there is a value for "PATH"
+## length(keys(org.Hs.eg.db, keytype="ENTREZID", column="PATH"))
+## is shorter than:
+## length(keys(org.Hs.eg.db, keytype="ENTREZID"))
+
+
+## debug(AnnotationDbi:::smartKeys)
+
+
+## Can just get keys that match a pattern
+## keys(org.Hs.eg.db, keytype="SYMBOL", pattern="BRCA")
+
+
+## Can get a key that matches a pattern on some other column
+## head(keys(org.Hs.eg.db,keytype="ENTREZID",pattern="MSX",column="SYMBOL"))
+
+
 
 
 
@@ -1037,7 +1367,7 @@ setMethod("keytypes", "ChipDb",
 )
 
 setMethod("keytypes", "GODb",
-    function(x) return("GOID") ## only one type makes sense
+    function(x) return(c("GOID","TERM","ONTOLOGY","DEFINITION")) ## only one type makes sense
 )
 
 
@@ -1067,7 +1397,7 @@ setMethod("keytypes", "GODb",
 
 
 ## debug(AnnotationDbi:::.queryForKeys)
-## debug(AnnotationDbi:::.makeKeytypeChoice)
+## debug(AnnotationDbi:::.keys)
 
 ## example of keys that uses keytype
 ## keys = keys(org.Hs.eg.db, keytype="ALIAS2EG")[1:4]
@@ -1096,7 +1426,7 @@ setMethod("keytypes", "GODb",
 #############################
 ## keytype example
 
-## library(hgu95av2.db); cols(hgu95av2.db); cols(org.Hs.eg.db); head(keys(org.Hs.eg.db, "ALIAS")); keys(org.Hs.eg.db, keytype="PROBEID")## should be an error
+## library(hgu95av2.db); columns(hgu95av2.db); columns(org.Hs.eg.db); head(keys(org.Hs.eg.db, "ALIAS")); keys(org.Hs.eg.db, keytype="PROBEID")## should be an error
 
 ## library(org.Hs.eg.db); keys2 = head(keys(org.Hs.eg.db, "ALIAS"));cols = c("SYMBOL", "GO");res <- select(org.Hs.eg.db, keys2, cols, keytype="ALIAS"); head(res); dim(res)
 
@@ -1166,7 +1496,7 @@ setMethod("keytypes", "GODb",
 ## foo = select(org.Hs.eg.db, keys=head(keys(org.Hs.eg.db),n=2),cols=c("ACCNUM")); head(foo)
 
 ## But this does NOT work (fixed):
-## foo = select(org.Hs.eg.db, keys=head(keys(org.Hs.eg.db),n=2),cols=head(cols(org.Hs.eg.db))); head(foo); head(cols(org.Hs.eg.db))
+## foo = select(org.Hs.eg.db, keys=head(keys(org.Hs.eg.db),n=2),cols=head(columns(org.Hs.eg.db))); head(foo); head(columns(org.Hs.eg.db))
 
 ## debug(AnnotationDbi:::.nameExceptions)
 ## debug(AnnotationDbi:::.addNAsInPlace)
